#Rver3.4.4 で動作確認済み

#ランダムパラメータが既知の場合の最尤推定
###########################################################################
#外部データの読み込み
data1 <- read.csv("学校データ.csv")

#レベル2の変数「プレテスト平均点」にCGMを施す
pre2.mdev <- data1$pre2.m-mean(data1$pre2.m)
#レベル2の変数「補習時間」にCGMを施す
time2.dev <- data1$time2-mean(data1$time2) 

#変数の表記をテキスト本文に合わせる
id <- data1$schoolID #学校ID
ytemp <- data1$post1 #目的変数ベクトルy
Xtemp <- cbind(1,data1$pre1.cwc) #レベル1データ行列X
Wtemp <- cbind(1,pre2.mdev,time2.dev)#レベル2データ行列W

#レベル2のランダムパラメータの設定
T <- matrix(c(29.393,0.413,0.413,0.077),ncol=2)

#レベル1のランダムパラメータの設定
sigma2 <- 84.916

y <- list()#yj
X <- list()#Xj
W <- list()#Wj
B <- list()#βj
V <- list()#Vj
D <- list()#Δj
WDW <- 0 #γの推定量の一部1
WDb <- 0 #γの推定量の一部２

for(j in 1:50)
{		

  #学校jのデータが収められている行番号を抽出
	s <- which(id==j)  
	y[[j]] <- ytemp[s] #yjを抽出
	X[[j]] <- Xtemp[s,] #Xjを抽出

  #Wjを作成(本文の行列表記を参照)
	W[[j]] <- rbind(c(Wtemp[s,][1,],0,0,0),c(0,0,0,Wtemp[s,][1,]))

  #Bjを算出しリストの要素に収める
	B[[j]] <- solve(t(X[[j]])%*%X[[j]])%*%t(X[[j]])%*%y[[j]] #Bjの算出と保存

  #Vjを算出しリストの要素に収める
	V[[j]] <- sigma2*solve(t(X[[j]])%*%X[[j]])　#Vjの算出と保存

  #Djを算出しリストの要素に収める
	D[[j]] <- T + V[[j]]	#Δjの算出と保存

  #γの推定量中の総和に関する処理
	WDW <- WDW + t(W[[j]])%*%solve(D[[j]])%*%W[[j]]
	WDb <- WDb + t(W[[j]])%*%solve(D[[j]])%*%B[[j]]
}	


gan <- t(solve(WDW)%*%WDb)
colnames(gan) <- c("g00","g01","g02","g10","g11","g12")
round(gan,3)

#WDWは上述のスクリプトで算出しているのでその逆行列の対角要素を求める
ganse <- diag(solve(WDW))
names(ganse) <- c("g00","g01","g02","g10","g11","g12")
round(sqrt(ganse),3)

#標準誤差と信頼区間
ll <- gan-1.96*sqrt(diag(solve(WDW)))
ul <- gan+1.96*sqrt(diag(solve(WDW)))
interval <- rbind(ll,ul)
rownames(interval) <- c("下限","上限")
interval
###########################################################################


#ランダム効果の推定
###########################################################################
L<-list()
for(j in 1:50)
{
	L[[j]] <- T%*%solve(D[[j]])	
}


Ba <- list()
u <- list()
I <- diag(2)

for(j in 1:50)
{
	Ba[[j]] <- L[[j]]%*%B[[j]]+(I-L[[j]])%*%W[[j]]%*%t(gan)	
	u[[j]] <- Ba[[j]]-W[[j]]%*%t(gan)	
}

u <- matrix(do.call("rbind",u),ncol=2,byrow=TRUE)
head(u,3)#最初3行を抽出


library(lmerTest)#パッケージlmerTestの読み込み
library(car)#パッケージcarの読み込み

#モデルの実行
crosslevel2 <- lmer(post1 ~ pre1.cwc +pre2.mdev + time2.dev
+ (pre1.cwc:pre2.mdev) + (pre1.cwc:time2.dev) 
+ (1 + pre1.cwc|schoolID),data=data1,REML=FALSE)

#ランダム効果の抽出
resran <- ranef(crosslevel2)$schoolID
head(resran,3)
###########################################################################



##ランダムパラメータが未知の場合の最尤推定(EMアルゴリズム)
###########################################################################
library(mvtnorm) #多変量正規分布の為のパッケージ
set.seed(123)
#学校別にデータを格納するための空のリストを作成
Af <- list()
Ar <- list()
AAf <- 0
Ay <- 0
for(j in 1:50)
{   
  Af[[j]] <- X[[j]]%*%W[[j]]
  Ar[[j]] <- X[[j]]
  AAf <- AAf + t(Af[[j]])%*%Af[[j]]
  Ay <- Ay + t(Af[[j]])%*%y[[j]]
}

I <- diag(100)#EMアルゴリズム内で必要となる単位行列

#thetarの初期値(tr0)
tr0 <- cbind(rnorm(50,0,1),rnorm(50,0,1))

AAt0 <- 0
#thetafの初期値(tf0)
for(j in 1:50)
{
  AAt0 <- t(Af[[j]])%*%Ar[[j]]%*%tr0[j,]
}
tf0 <- solve(AAf)%*%(Ay-AAt0)

#Tの初期値(T0)
tt0 <- 0
for(j in 1:50)
{
  tt0 <- tt0 + tr0[j,]%*%t(tr0[j,])
}	
T0 <- tt0/50


#sigma2の初期値(s0)
rr0 <- 0
for(j in 1:50)
{	
  rr0 <- rr0 + t((y[[j]]-Af[[j]]%*%tf0-Ar[[j]]%*%tr0[j,]))%*%(y[[j]]
  -Af[[j]]%*%tf0-Ar[[j]]%*%tr0[j,])
}
s0 <- rr0/5000


#---------------
#EMアルゴリズムスタート
repeat #EMアルゴリズムのループ開始
{
	
#E-step
   #Cの計算
   C <- list()
   for(j in 1:50)
   {
   　C[[j]] <- t(Ar[[j]])%*%Ar[[j]] + as.numeric(s0)*solve(T0)
   }
	
   #t.astの計算
   t.ast.pre <- list()
   for(j in 1:50)
   {
      t.ast.pre[[j]] <- solve(C[[j]])%*%t(Ar[[j]])%*%(y[[j]]-Af[[j]]%*%tf0)
   }
	
	t.ast <- t(do.call("cbind",t.ast.pre))

	
	#CDSS1の計算
	cds1 <- 0
	for(j in 1:50)
	{	
		cds1 <- cds1 +t(Af[[j]])%*%Ar[[j]]%*%t.ast[j,]
	}
	
	#CDSS2の計算
	cds2 <- 0
	for(j in 1:50)
	{	
		cds2 <- cds2 + t.ast[j,]%*%t(t.ast[j,]) + as.numeric(s0)*solve(C[[j]])
	}
	
	#CDSS3の計算
	rr <- 0
	stat <- 0
	for(j in 1:50)
	{	
		rr <- rr + t((y[[j]]-Af[[j]]%*%tf0-Ar[[j]]%*%t.ast[j,]))%*%(y[[j]]
		-Af[[j]]%*%tf0-Ar[[j]]%*%t.ast[j,])	
		stat <- stat + solve(C[[j]])%*%t(Ar[[j]])%*%Ar[[j]]
	}
		
	cds3 <- rr + s0*sum(diag(stat)) 	

	#M-step
	tf1 <- solve(AAf)%*%(Ay-cds1)
	T1 <- cds2/50
	s1 <- cds3/5000

	
	#前回の対数尤度
	ll0 <- 0
	for(j in 1:50)
	{
		
		mu.vec <- c(Af[[j]]%*%tf0,rep(0,2))
		var1 <- Ar[[j]]%*%T0%*%t(Ar[[j]])+as.numeric(s0)*I
		cov1 <- Ar[[j]]%*%T0
		cov2 <- T0%*%t(Ar[[j]])
		var2 <- T0			
		COV <- rbind(cbind(var1,cov1),cbind(cov2,var2))			
		ll0 <- ll0 + log(dmvnorm(c(y[[j]],tr0[j,]),mu.vec,COV))
	}
	
	#今回の対数尤度
	ll1 <- 0
	for(j in 1:50)
	{
		mu.vec <- c(Af[[j]]%*%tf1,rep(0,2))
		var1 <- Ar[[j]]%*%T1%*%t(Ar[[j]])+as.numeric(s1)*I
		cov1 <- Ar[[j]]%*%T1
		cov2 <- T1%*%t(Ar[[j]])
		var2 <- T1	
		COV <- rbind(cbind(var1,cov1),cbind(cov2,var2))	
		ll1 <- ll1 + log(dmvnorm(c(y[[j]],t.ast[j,]),mu.vec,COV))
	}
	
	dif <-abs(ll0-ll1)
	print(dif)
		
	if(dif<0.05)
	{
		rownames(T1) <- colnames(T1) <- c("β0","β1")		
		rownames(tf1) <- c("g00","g01","g02","g10","g11","g12")
		colnames(s1) <- "sigma2"
		colnames(t.ast) <- c("u0","u1")
		break
	}	
	
	T0 <- T1
	tf0 <- tf1
	s0 <- s1
	tr0 <- t.ast
}

round(T1,3)
round(t(tf1),3)
round(s1,3)
##########################################################################


########################################################
library(rstan)#スタンの読み込み

data1 <-read.csv("学校データ.csv")

N <- 5000 #生徒数
J <- 50 #学校数
schoolID <- data1$schoolID #学校ID

pre1cwc <- data1$pre1.cwc #プレテスト得点(CWC)
post1 <- data1$post1 #ポストテスト得点

loc <- seq(1,5000,100) #各学校の一人目の生徒の位置を取得

#各学校のプレテスト平均(CGM)
pre2mdev <- (data1$pre2.m-mean(data1$pre2.m))[loc] 

#各学校の自習時間(CGM)
time2dev <- (data1$time2-mean(data1$time2))[loc] 

#stanに受け渡すオブジェクト名の一覧を作成
datlist <- c("N", "J", "post1", "pre1cwc","schoolID","pre2mdev","time2dev")

#stanの実行
resstan <- stan(file="multilevel.stan", data = datlist,chains=3, iter=10000,warmup=
5000,seed=123)

#収束判定，点推定値，確信区間の算出
print(resstan,pars=c("g00","g01","g02","g10","g11","g12","sigma","T"),digits=3)

#カーネル密度関数の重ね書き
stan_dens(resstan,pars="T[1,1]",separate_chains=TRUE) 

#ＭＣＭＣサンプルの自己相関プロット
stan_ac(resstan,pars="T[1,1]",separate_chains=TRUE)
########################################################